from typing import MutableSequence, Sequence, Tuple, Union

# note: this shouldn't be necessary, but apparently is
Number = Union[int, float]
ImageSequence = Sequence[Number]
MutableImageSequence = MutableSequence[Number]
RGBTuple = Union[Tuple[Number, Number, Number], Sequence[Number]]



from pathlib import Path
from typing import Dict

import pytest
from PIL import Image
from pixelmatch.contrib import PIL

from pixelmatch import pixelmatch

FIXTURES_PATH = Path(__file__).parent / "fixtures"
OPTIONS = {"threshold": 0.05}


def read_img(name: str):
    return Image.open(FIXTURES_PATH / f"{name}.png")


def pil_to_flatten_data(img):
    """
    Convert data from [(R1, G1, B1, A1), (R2, G2, B2, A2)] to [R1, G1, B1, A1, R2, G2, B2, A2]
    """
    return [x for p in img.convert("RGBA").getdata() for x in p]


testdata = [
    ["1a", "1b", "1diff", OPTIONS, 143],
    [
        "1a",
        "1b",
        "1diffmask",
        {"threshold": 0.05, "includeAA": False, "diff_mask": True},
        143,
    ],
    ["1a", "1a", "1emptydiffmask", {"threshold": 0, "diff_mask": True}, 0],
    [
        "2a",
        "2b",
        "2diff",
        {
            "threshold": 0.05,
            "alpha": 0.5,
            "aa_color": [0, 192, 0],
            "diff_color": [255, 0, 255],
        },
        12437,
    ],
    ["3a", "3b", "3diff", OPTIONS, 212],
    ["4a", "4b", "4diff", OPTIONS, 36049],
    ["5a", "5b", "5diff", OPTIONS, 0],
    ["6a", "6b", "6diff", OPTIONS, 51],
    ["6a", "6a", "6empty", {"threshold": 0}, 0],
    ["7a", "7b", "7diff", OPTIONS, 9856],
]


@pytest.mark.parametrize(
    "img_path_1,img_path_2,diff_path,options,expected_mismatch", testdata
)
def test_pixelmatch(
    img_path_1: str,
    img_path_2: str,
    diff_path: str,
    options: Dict,
    expected_mismatch: int,
    benchmark,
):

    img1 = read_img(img_path_1)
    img2 = read_img(img_path_2)
    width, height = img1.size
    img1_data = pil_to_flatten_data(img1)
    img2_data = pil_to_flatten_data(img2)
    diff_data = [0.0] * len(img1_data)

    mismatch = benchmark(
        pixelmatch, img1_data, img2_data, width, height, diff_data, **options
    )
    mismatch2 = pixelmatch(img1_data, img2_data, width, height, None, **options)

    expected_diff = read_img(diff_path)
    assert diff_data == pil_to_flatten_data(expected_diff), "diff image"
    assert mismatch == expected_mismatch, "number of mismatched pixels"
    assert mismatch == mismatch2, "number of mismatched pixels without diff"


@pytest.mark.parametrize(
    "img_path_1,img_path_2,diff_path,options,expected_mismatch", testdata
)
def test_pixelmatch_failfast(
    img_path_1: str,
    img_path_2: str,
    diff_path: str,
    options: Dict,
    expected_mismatch: int,
    benchmark,
):
    img1 = read_img(img_path_1)
    img2 = read_img(img_path_2)
    width, height = img1.size
    img1_data = pil_to_flatten_data(img1)
    img2_data = pil_to_flatten_data(img2)
    if expected_mismatch:
        assert (
            benchmark(
                pixelmatch,
                img1_data,
                img2_data,
                width,
                height,
                fail_fast=True,
                **options,
            )
            == 1
        )


@pytest.mark.parametrize(
    "img_path_1,img_path_2,diff_path,options,expected_mismatch", testdata
)
def test_PIL_pixelmatch(
    img_path_1: str,
    img_path_2: str,
    diff_path: str,
    options: Dict,
    expected_mismatch: int,
    benchmark,
):
    img1 = read_img(img_path_1)
    img2 = read_img(img_path_2)
    diff_data = Image.new("RGBA", img1.size)

    mismatch = benchmark(PIL.pixelmatch, img1, img2, diff_data, **options)
    mismatch2 = PIL.pixelmatch(img1, img2, **options)

    assert pil_to_flatten_data(diff_data) == pil_to_flatten_data(
        read_img(diff_path)
    ), "diff image"
    assert mismatch == expected_mismatch, "number of mismatched pixels"
    assert mismatch == mismatch2, "number of mismatched pixels without diff"



"""Functions to facilitate direct comparison of PIL.Image instances"""
from typing import List, Optional, Tuple

from PIL.Image import Image

from pixelmatch import core
from pixelmatch.types import ImageSequence, MutableImageSequence, RGBTuple


def pixelmatch(
    img1: Image,
    img2: Image,
    output: Optional[Image] = None,
    threshold: float = 0.1,
    includeAA: bool = False,
    alpha: float = 0.1,
    aa_color: RGBTuple = (255, 255, 0),
    diff_color: RGBTuple = (255, 0, 0),
    diff_mask: bool = False,
    fail_fast: bool = False,
) -> int:
    """
    Compares two images, writes the output diff and returns the number of mismatched pixels.
    Serves the same purpose as pixelmatch.pixelmatch, but takes PIL.Images as input instead
    of raw image data.

    :param img1: PIL.Image data to compare with img2. Must be the same size as img2
    :param img2: PIL.Image data to compare with img2. Must be the same size as img1
    :param output: Image data to write the diff to. Should be the same size as
    :param threshold: matching threshold (0 to 1); smaller is more sensitive, defaults to 1
    :param includeAA: whether or not to skip anti-aliasing detection, ie if includeAA is True,
        detecting and ignoring anti-aliased pixels is disabled. Defaults to False
    :param alpha: opacity of original image in diff output, defaults to 0.1
    :param aa_color: tuple of RGB color of anti-aliased pixels in diff output,
        defaults to (255, 255, 0) (yellow)
    :param diff_color: tuple of RGB color of the color of different pixels in diff output,
        defaults to (255, 0, 0) (red)
    :param diff_mask: whether or not to draw the diff over a transparent background (a mask),
        defaults to False
    :param fail_fast: if true, will return after first different pixel. Defaults to false
    :return: number of pixels that are different or 1 if fail_fast == true
    """
    width, height = img1.size
    img1 = from_PIL_to_raw_data(img1)
    img2 = from_PIL_to_raw_data(img2)

    if output is not None:
        raw_output: Optional[MutableImageSequence] = from_PIL_to_raw_data(output)
    else:
        raw_output = None

    diff_pixels = core.pixelmatch(
        img1,
        img2,
        width,
        height,
        raw_output,
        threshold=threshold,
        includeAA=includeAA,
        alpha=alpha,
        aa_color=aa_color,
        diff_color=diff_color,
        diff_mask=diff_mask,
        fail_fast=fail_fast,
    )

    if raw_output is not None and output is not None:
        output.putdata(to_PIL_from_raw_data(raw_output))

    return diff_pixels


def from_PIL_to_raw_data(pil_img: Image) -> MutableImageSequence:
    """
    Converts a PIL.Image object from [(R1, B1, A1, A1), (R2, ...), ...] to our raw data format
    [R1, G1, B1, A1, R2, ...].

    :param pil_img:
    :return:
    """
    return [item for sublist in pil_img.convert("RGBA").getdata() for item in sublist]


def to_PIL_from_raw_data(
    raw_data: ImageSequence,
) -> List[Tuple[float, float, float, float]]:
    """
    Converts from the internal raw data format of [R1, G1, B1, A1, R2, ...] to PIL's raw data format, ie
    [(R1, B1, A1, A1), (R2, ...), ...]
    :return: raw image data in a PIL appropriate format
    """
    return [*zip(raw_data[::4], raw_data[1::4], raw_data[2::4], raw_data[3::4])]



from .types import ImageSequence, MutableImageSequence


def antialiased(
    img: ImageSequence, x1: int, y1: int, width: int, height: int, img2: ImageSequence
) -> bool:
    """
    check if a pixel is likely a part of anti-aliasing;
    based on "Anti-aliased Pixel and Intensity Slope Detector" paper by V. Vysniauskas, 2009
    """
    x0 = max(x1 - 1, 0)
    y0 = max(y1 - 1, 0)
    x2 = min(x1 + 1, width - 1)
    y2 = min(y1 + 1, height - 1)
    pos = (y1 * width + x1) * 4
    zeroes = int(x1 == x0 or x1 == x2 or y1 == y0 or y1 == y2)
    min_delta = max_delta = 0.0
    min_x = min_y = max_x = max_y = 0

    # go through 8 adjacent pixels
    for x in range(x0, x2 + 1):
        for y in range(y0, y2 + 1):
            if x == x1 and y == y1:
                continue

            # brightness delta between the center pixel and adjacent one
            delta = color_delta(img, img, pos, (y * width + x) * 4, True)

            # count the number of equal, darker and brighter adjacent pixels
            if delta == 0:
                zeroes += 1
                # if found more than 2 equal siblings, it's definitely not anti-aliasing
                if zeroes > 2:
                    return False

            # remember the darkest pixel
            elif delta < min_delta:
                min_delta = delta
                min_x = x
                min_y = y

            # remember the brightest pixel
            elif delta > max_delta:
                max_delta = delta
                max_x = x
                max_y = y

    # if there are no both darker and brighter pixels among siblings, it's not anti-aliasing
    if min_delta == 0 or max_delta == 0:
        return False

    # if either the darkest or the brightest pixel has 3+ equal siblings in both images
    # (definitely not anti-aliased), this pixel is anti-aliased
    return (
        has_many_siblings(img, min_x, min_y, width, height)
        and has_many_siblings(img2, min_x, min_y, width, height)
    ) or (
        has_many_siblings(img, max_x, max_y, width, height)
        and has_many_siblings(img2, max_x, max_y, width, height)
    )


def has_many_siblings(
    img: ImageSequence, x1: int, y1: int, width: int, height: int
) -> bool:
    """
    check if a pixel has 3+ adjacent pixels of the same color.
    """
    x0 = max(x1 - 1, 0)
    y0 = max(y1 - 1, 0)
    x2 = min(x1 + 1, width - 1)
    y2 = min(y1 + 1, height - 1)
    pos = (y1 * width + x1) * 4
    zeroes = int(x1 == x0 or x1 == x2 or y1 == y0 or y1 == y2)

    # go through 8 adjacent pixels
    for x in range(x0, x2 + 1):
        for y in range(y0, y2 + 1):
            if x == x1 and y == y1:
                continue

            pos2 = (y * width + x) * 4
            if all(img[pos + offset] == img[pos2 + offset] for offset in range(4)):
                zeroes += 1

            if zeroes > 2:
                return True

    return False


def color_delta(
    img1: ImageSequence, img2: ImageSequence, k: int, m: int, y_only: bool = False
) -> float:
    """
    calculate color difference according to the paper "Measuring perceived color difference
    using YIQ NTSC transmission color space in mobile applications" by Y. Kotsarenko and F. Ramos
    """
    r1, g1, b1, a1 = [img1[k + offset] for offset in range(4)]
    r2, g2, b2, a2 = [img2[m + offset] for offset in range(4)]

    if a1 == a2 and r1 == r2 and g1 == g2 and b1 == b2:
        return 0.0

    if a1 < 255:
        a1 /= 255
        r1, b1, g1 = blendRGB(r1, b1, g1, a1)

    if a2 < 255:
        a2 /= 255
        r2, b2, g2 = blendRGB(r2, b2, g2, a2)

    y = rgb2y(r1, g1, b1) - rgb2y(r2, g2, b2)

    if y_only:
        # brightness difference only
        return y

    i = rgb2i(r1, g1, b1) - rgb2i(r2, g2, b2)
    q = rgb2q(r1, g1, b1) - rgb2q(r2, g2, b2)

    return 0.5053 * y * y + 0.299 * i * i + 0.1957 * q * q


def rgb2y(r: float, g: float, b: float) -> float:
    return r * 0.29889531 + g * 0.58662247 + b * 0.11448223


def rgb2i(r: float, g: float, b: float) -> float:
    return r * 0.59597799 - g * 0.27417610 - b * 0.32180189


def rgb2q(r: float, g: float, b: float) -> float:
    return r * 0.21147017 - g * 0.52261711 + b * 0.31114694


def blendRGB(r: float, g: float, b: float, a: float):
    """
    Blend r, g, and b with a
    :param r: red channel to blend with a
    :param g: green channel to blend with a
    :param b: blue channel to blend with a
    :param a: alpha to blend with
    :return: tuple of blended r, g, b
    """
    return blend(r, a), blend(g, a), blend(b, a)


def blend(c: float, a: float) -> float:
    """blend semi-transparent color with white"""
    return 255 + (c - 255) * a


def draw_pixel(
    output: MutableImageSequence, pos: int, r: float, g: float, b: float
) -> None:
    output[pos + 0] = int(r)
    output[pos + 1] = int(g)
    output[pos + 2] = int(b)
    output[pos + 3] = 255


def draw_gray_pixel(
    img: ImageSequence, i: int, alpha: float, output: MutableImageSequence
) -> None:
    r = img[i + 0]
    g = img[i + 1]
    b = img[i + 2]
    val = blend(rgb2y(r, g, b), alpha * img[i + 3] / 255)
    draw_pixel(output, i, val, val, val)



from typing import Optional

from .types import ImageSequence, MutableImageSequence, RGBTuple
from .utils import antialiased, color_delta, draw_gray_pixel, draw_pixel


def pixelmatch(
    img1: ImageSequence,
    img2: ImageSequence,
    width: int,
    height: int,
    output: Optional[MutableImageSequence] = None,
    threshold: float = 0.1,
    includeAA: bool = False,
    alpha: float = 0.1,
    aa_color: RGBTuple = (255, 255, 0),
    diff_color: RGBTuple = (255, 0, 0),
    diff_mask: bool = False,
    fail_fast: bool = False,
) -> int:
    """
    Compares two images, writes the output diff and returns the number of mismatched pixels.
    'Raw image data' refers to a 1D, indexable collection of image data in the
    format [R1, G1, B1, A1, R2, G2, ...].

    :param img1: Image data to compare with img2. Must be the same size as img2
    :param img2: Image data to compare with img2. Must be the same size as img1
    :param width: Width of both images (they should be the same).
    :param height: Height of both images (they should be the same).
    :param output: Image data to write the diff to. Should be the same size as
    :param threshold: matching threshold (0 to 1); smaller is more sensitive, defaults to 1
    :param includeAA: whether or not to skip anti-aliasing detection, ie if includeAA is True,
        detecting and ignoring anti-aliased pixels is disabled. Defaults to False
    :param alpha: opacity of original image in diff output, defaults to 0.1
    :param aa_color: tuple of RGB color of anti-aliased pixels in diff output,
        defaults to (255, 255, 0) (yellow)
    :param diff_color: tuple of RGB color of the color of different pixels in diff output,
        defaults to (255, 0, 0) (red)
    :param diff_mask: whether or not to draw the diff over a transparent background (a mask),
        defaults to False
    :param fail_fast: if true, will return after first different pixel. Defaults to false
    :return: number of pixels that are different or 1 if fail_fast == true
    """

    if len(img1) != len(img2):
        raise ValueError("Image sizes do not match.", len(img1), len(img2))
    if output and len(output) != len(img1):
        raise ValueError(
            "Diff image size does not match img1 & img2.", len(img1), len(output)
        )

    if len(img1) != width * height * 4:
        raise ValueError(
            "Image data size does not match width/height.",
            len(img1),
            width * height * 4,
        )

    # fast path if identical
    if img1 == img2:
        if output and not diff_mask:
            for i in range(width * height):
                draw_gray_pixel(img1, 4 * i, alpha, output)

        return 0

    # maximum acceptable square distance between two colors;
    # 35215 is the maximum possible value for the YIQ difference metric
    maxDelta = 35215 * threshold * threshold

    diff = 0
    aaR, aaG, aaB = aa_color
    diffR, diffG, diffB = diff_color

    # compare each pixel of one image against the other one
    for y in range(height):
        for x in range(width):
            pos = (y * width + x) * 4

            # squared YUV distance between colors at this pixel position
            delta = color_delta(img1, img2, pos, pos)

            # the color difference is above the threshold
            if delta > maxDelta:
                # check it's a real rendering difference or just anti-aliasing
                if not includeAA and (
                    antialiased(img1, x, y, width, height, img2)
                    or antialiased(img2, x, y, width, height, img1)
                ):
                    # one of the pixels is anti-aliasing; draw as yellow and do not count as difference
                    # note that we do not include such pixels in a mask
                    if output and not diff_mask:
                        draw_pixel(output, pos, aaR, aaG, aaB)
                else:
                    # found substantial difference not caused by anti-aliasing; draw it as red
                    if output:
                        draw_pixel(output, pos, diffR, diffG, diffB)
                    if fail_fast:
                        return 1
                    diff += 1

            elif output:
                # pixels are similar; draw background as grayscale image blended with white
                if not diff_mask:
                    draw_gray_pixel(img1, pos, alpha, output)

    # return the number of different pixels
    return diff
